# Copyright 2018-2025 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import collections
import datetime
import logging
from abc import abstractmethod
from enum import Enum
from typing import Any, Dict, List, Optional, Sequence, Tuple

import attr
from openlineage.client.event_v2 import Dataset, InputDataset, Job, OutputDataset, Run, RunEvent, RunState
from openlineage.client.facet_v2 import (
    DatasetFacet,
    InputDatasetFacet,
    JobFacet,
    OutputDatasetFacet,
    RunFacet,
    column_lineage_dataset,
    data_quality_assertions_dataset,
    datasource_dataset,
    documentation_dataset,
    external_query_run,
    job_type_job,
    output_statistics_output_dataset,
    processing_engine_run,
    schema_dataset,
    sql_job,
    tags_run,
)
from openlineage.client.uuid import generate_new_uuid
from openlineage.common.provider.dbt.facets import DbtRunRunFacet, DbtVersionRunFacet, ParentRunMetadata
from openlineage.common.provider.dbt.utils import __version__ as openlineage_version
from openlineage.common.provider.snowflake import fix_account_name
from openlineage.common.utils import get_from_multiple_chains, get_from_nullable_chain
from openlineage_sql import parse as parse_sql


class Adapter(Enum):
    # This class represents supported adapters.
    BIGQUERY = "bigquery"
    SNOWFLAKE = "snowflake"
    REDSHIFT = "redshift"
    SPARK = "spark"
    POSTGRES = "postgres"
    DATABRICKS = "databricks"
    SQLSERVER = "sqlserver"
    DREMIO = "dremio"
    ATHENA = "athena"
    DUCKDB = "duckdb"
    TRINO = "trino"
    GLUE = "glue"
    CLICKHOUSE = "clickhouse"

    @staticmethod
    def adapters() -> str:
        # String representation of all supported adapter names
        return ",".join([f"`{x.value}`" for x in list(Adapter)])


class SparkConnectionMethod(Enum):
    THRIFT = "thrift"
    ODBC = "odbc"
    HTTP = "http"

    @staticmethod
    def methods():
        return [x.value for x in SparkConnectionMethod]


class UnsupportedDbtCommand(Exception):
    pass


@attr.define
class ModelNode:
    # in reality Literal["model", "test", "snapshot", "source", "seed"]
    # but way too much work to convince mypy that it's true
    type: str
    metadata_node: Dict
    catalog_node: Optional[Dict] = None


@attr.define
class DbtRun:
    started_at: str
    completed_at: str
    status: str
    inputs: List[Dataset]
    output: Optional[Dataset]
    job_name: str
    namespace: str
    run_id: str = attr.field(factory=lambda: str(generate_new_uuid()))


@attr.define
class DbtRunResult:
    start: RunEvent
    complete: Optional[RunEvent] = None
    fail: Optional[RunEvent] = None


@attr.define
class DbtEvents:
    starts: List[RunEvent] = attr.field(factory=list)
    completes: List[RunEvent] = attr.field(factory=list)
    fails: List[RunEvent] = attr.field(factory=list)

    def events(self):
        return self.starts + self.completes + self.fails

    def add(self, item: Optional[DbtRunResult]):
        if not item:
            return
        self.starts.append(item.start)
        if item.complete:
            self.completes.append(item.complete)
        if item.fail:
            self.fails.append(item.fail)

    def __iadd__(self, other):
        if isinstance(other, DbtEvents):
            self.starts += other.starts
            self.completes += other.completes
            self.fails += other.fails
            return self
        raise NotImplementedError


@attr.define
class DbtRunContext:
    manifest: Dict
    run_results: Dict
    catalog: Optional[Dict] = None


class DbtArtifactProcessor:
    should_raise_on_unsupported_command = True

    def __init__(
        self,
        producer: str,
        job_namespace: str,
        skip_errors: bool = False,
        logger: Optional[logging.Logger] = None,
        models: Optional[Sequence[str]] = None,
        selector: Optional[str] = None,
        openlineage_job_name: Optional[str] = None,
    ):
        self.producer = producer
        self._dbt_run_metadata: Optional[ParentRunMetadata] = None
        self.logger = logger or logging.getLogger(f"{self.__class__.__module__}.{self.__class__.__name__}")

        self.openlineage_job_name = openlineage_job_name
        self.job_namespace = job_namespace
        self.dataset_namespace = ""
        self.skip_errors = skip_errors
        self.run_metadata: dict[str, Any] = {}
        self.command = None
        self.models = models or []
        self.selector = selector
        self.manifest_version = None
        self.adapter_type: Adapter | None = None

    @property
    def dbt_run_metadata(self):
        return self._dbt_run_metadata

    @dbt_run_metadata.setter
    def dbt_run_metadata(self, metadata: ParentRunMetadata):
        self._dbt_run_metadata = metadata

    @abstractmethod
    def get_dbt_metadata(self):
        ...

    def parse(self) -> DbtEvents:
        """
        Parse dbt manifest and run_result and produce OpenLineage events.
        """
        manifest, run_result, profile, catalog = self.get_dbt_metadata()
        self.manifest_version = self.get_schema_version(manifest)

        self.run_metadata = run_result["metadata"]
        self.command = run_result["args"]["which"]

        self.extract_adapter_type(profile)
        self.extract_dataset_namespace(profile)

        nodes = {}
        # Filter nodes
        for name, node in manifest["nodes"].items():
            if any(name.startswith(prefix) for prefix in ("model.", "test.", "snapshot.")):
                nodes[name] = node

        context = DbtRunContext(manifest, run_result, catalog)

        events = DbtEvents()

        if self.command not in ["run", "build", "test", "seed", "snapshot"]:
            if self.should_raise_on_unsupported_command:
                raise UnsupportedDbtCommand(
                    f"Not recognized run command {self.command} - "
                    "should be run, build, test, seed or snapshot"
                )
            else:
                return events

        if self.command in ["run", "build", "seed", "snapshot"]:
            events += self.parse_execution(context, nodes)
        if self.command in ["test", "build"]:
            events += self.parse_test(context, nodes)
        return events

    @classmethod
    def get_schema_version(cls, metadata):
        str_schema_version = get_from_nullable_chain(metadata, ["metadata", "dbt_schema_version"])
        return cls.get_version_number(str_schema_version)

    def get_query_id(self, run_result: dict[str, Any]) -> Optional[str]:
        # Validate that there is an adapter_response in the run_result
        if "adapter_response" not in run_result:
            return None

        # Default to query_id for all Adapters
        query_id_key: str = "query_id"

        # Use the adapter type to make sure the correct key is used
        if self.adapter_type == Adapter.BIGQUERY:
            query_id_key = "job_id"

        query_id: Optional[str] = run_result["adapter_response"].get(query_id_key)

        if isinstance(query_id, str):
            # For Databricks, "N/A" could be returned if the query_id is None; catch that
            return None if query_id.lower() == "n/a" else query_id

        return query_id

    @staticmethod
    def get_version_number(version: str) -> int:
        # "https://schemas.getdbt.com/dbt/manifest/v2.json" -> "v2.json"
        file = version.split("/")[-1]
        # "v2.json" -> 2
        return int(file.split(".")[0][1:])

    def parse_execution(self, context: DbtRunContext, nodes: Dict) -> DbtEvents:
        events = DbtEvents()
        for run in context.run_results["results"]:
            name = run["unique_id"]

            # Pull the available query_id from the run_results
            query_id: str | None = self.get_query_id(run)

            if not any(name.startswith(prefix) for prefix in ("model.", "source.", "snapshot.")):
                continue
            if run["status"] == "skipped":
                continue

            output_node = nodes[name]
            started_at, completed_at = self.get_timings(run["timing"])

            inputs = []
            for node in context.manifest["parent_map"][run["unique_id"]]:
                if node.startswith("model."):
                    inputs.append(
                        ModelNode(
                            type="model",
                            metadata_node=nodes[node],
                            catalog_node=get_from_nullable_chain(context.catalog, ["nodes", node]),
                        )
                    )
                elif node.startswith("source."):
                    inputs.append(
                        ModelNode(
                            type="source",
                            metadata_node=context.manifest["sources"][node],
                            catalog_node=get_from_nullable_chain(context.catalog, ["sources", node]),
                        )
                    )

            run_id = str(generate_new_uuid())
            if name.startswith("snapshot."):
                jobType = "SNAPSHOT"
                job_name = self._format_dataset_name(
                    output_node["database"],
                    output_node["schema"],
                    self.removeprefix(run["unique_id"], "snapshot."),
                ) + (".build.snapshot" if self.command == "build" else ".snapshot")
            else:
                jobType = "MODEL"
                job_name = self._format_dataset_name(
                    output_node["database"],
                    output_node["schema"],
                    self.removeprefix(run["unique_id"], "model."),
                ) + (".build.run" if self.command == "build" else "")

            if self.manifest_version >= 7:  # type: ignore
                sql = output_node.get("compiled_code", None)
            else:
                sql = output_node["compiled_sql"]

            job_facets: Dict[str, JobFacet] = {
                "jobType": job_type_job.JobTypeJobFacet(
                    jobType=jobType,
                    integration="DBT",
                    processingType="BATCH",
                    producer=self.producer,
                )
            }
            if sql:
                job_facets["sql"] = sql_job.SQLJobFacet(query=sql, dialect=self.extract_dialect())

            run_facets: Dict[str, RunFacet] = {}
            if tags := output_node.get("tags", None):
                run_facets["tags"] = tags_run.TagsRunFacet(
                    tags=[tags_run.TagsRunFacetFields(key=tag, value="true", source="DBT") for tag in tags]
                )

            output_dataset = self.node_to_output_dataset(
                ModelNode(
                    type=jobType.lower(),
                    metadata_node=output_node,
                    catalog_node=get_from_nullable_chain(context.catalog, ["nodes", run["unique_id"]]),
                ),
                has_facets=True,
                adapter_response=run.get("adapter_response", None),
            )

            # Add column lineage if SQL is available
            if sql:
                column_lineage = self.get_column_lineage(output_dataset.namespace, sql)
                if column_lineage:
                    output_dataset.facets["columnLineage"] = column_lineage  # type: ignore

            events.add(
                self.to_openlineage_events(
                    run["status"],
                    started_at,
                    completed_at,
                    self.get_run(run_id=run_id, query_id=query_id, run_facets=run_facets),
                    Job(namespace=self.job_namespace, name=job_name, facets=job_facets),
                    [self.node_to_dataset(node, has_facets=True) for node in inputs],
                    output_dataset,
                )
            )
        return events

    def parse_test(self, context: DbtRunContext, nodes: Dict) -> DbtEvents:
        # The tests can have different timings, so just take current time
        started_at = datetime.datetime.now(datetime.timezone.utc).isoformat()
        completed_at = datetime.datetime.now(datetime.timezone.utc).isoformat()

        assertions = self.parse_assertions(context, nodes)

        events = DbtEvents()
        manifest_nodes = {**context.manifest["nodes"], **context.manifest["sources"]}
        for name, node in manifest_nodes.items():
            if name.startswith("model."):
                node_type = "model"
            elif name.startswith("source."):
                node_type = "source"
            else:
                continue
            if len(assertions[name]) == 0:
                continue

            assertion_facet = data_quality_assertions_dataset.DataQualityAssertionsDatasetFacet(
                assertions=assertions[name]
            )

            namespace, name, _, _ = self.extract_dataset_data(
                ModelNode(type=node_type, metadata_node=node), assertion_facet, has_facets=False
            )

            job_name = self._format_dataset_name(
                node["database"],
                node["schema"],
                self.removeprefix(node["unique_id"], "model."),
            ) + (".build.test" if self.command == "build" else ".test")

            job_facets: Dict[str, JobFacet] = {
                "jobType": job_type_job.JobTypeJobFacet(
                    jobType="TEST",
                    integration="DBT",
                    processingType="BATCH",
                    producer=self.producer,
                )
            }

            run_facets: Dict[str, RunFacet] = {}
            if tags := node.get("tags", None):
                run_facets["tags"] = tags_run.TagsRunFacet(
                    tags=[tags_run.TagsRunFacetFields(key=tag, value="true", source="DBT") for tag in tags]
                )

            run_id = str(generate_new_uuid())
            dataset_facets: Dict[str, InputDatasetFacet] = {"dataQualityAssertions": assertion_facet}
            events.add(
                self.to_openlineage_events(
                    "success",
                    started_at,
                    completed_at,
                    self.get_run(run_id=run_id, run_facets=run_facets),
                    Job(namespace=self.job_namespace, name=job_name, facets=job_facets),
                    [
                        InputDataset(
                            namespace=namespace,
                            name=name,
                            inputFacets=dataset_facets,
                            # TODO: remove this next release
                            facets=dataset_facets,  # type: ignore
                        )
                    ],
                    None,
                )
            )
        return events

    def parse_assertions(
        self, context: DbtRunContext, nodes: Dict
    ) -> Dict[str, List[data_quality_assertions_dataset.Assertion]]:
        assertions = collections.defaultdict(list)
        for run in context.run_results["results"]:
            if not run["unique_id"].startswith("test."):
                continue
            test_node = nodes[run["unique_id"]]

            model_node = None
            for node in context.manifest["parent_map"][run["unique_id"]]:
                if any(node.startswith(prefix) for prefix in ["model.", "source.", "seed."]):
                    model_node = node

            if self.manifest_version >= 12:  # type: ignore
                name = test_node["name"]
                node_columns = test_node

            else:
                name = test_node["test_metadata"]["name"]
                node_columns = test_node["test_metadata"]

            assertions[model_node].append(
                data_quality_assertions_dataset.Assertion(
                    assertion=name,
                    success=True if run["status"] == "pass" else False,
                    column=get_from_nullable_chain(node_columns, ["kwargs", "column_name"]),
                )
            )

            if not model_node:
                self.logger.warning(f"Model node connected to test {nodes[run['unique_id']]} not found")
        return assertions  # type: ignore

    def to_openlineage_events(self, *args, **kwargs) -> Optional[DbtRunResult]:
        try:
            return self._to_openlineage_events(*args, **kwargs)
        except Exception as e:
            if self.skip_errors:
                return None
            raise ValueError from e

    def _to_openlineage_events(
        self,
        status: str,
        started_at: str,
        completed_at: str,
        run: Run,
        job: Job,
        inputs: List[InputDataset],
        output: Optional[OutputDataset],
    ) -> Optional[DbtRunResult]:
        start = RunEvent(
            eventType=RunState.START,
            eventTime=started_at,
            run=run,
            job=job,
            producer=self.producer,
            inputs=inputs,
            outputs=[output] if output else [],
        )
        if status == "success":
            return DbtRunResult(
                start,
                complete=RunEvent(
                    eventType=RunState.COMPLETE,
                    eventTime=completed_at,
                    run=run,
                    job=job,
                    producer=self.producer,
                    inputs=inputs,
                    outputs=[output] if output else [],
                ),
            )
        elif status == "error":
            return DbtRunResult(
                start,
                fail=RunEvent(
                    eventType=RunState.FAIL,
                    eventTime=completed_at,
                    run=run,
                    job=job,
                    producer=self.producer,
                    inputs=inputs,
                    outputs=[],
                ),
            )
        else:
            # Should not happen?
            raise ValueError(f"Run status was {status}, should be in ['success', 'skipped', 'error']")

    def node_to_dataset(
        self,
        node: ModelNode,
        has_facets: bool = False,
    ) -> Dataset:
        namespace, name, facets, _ = self.extract_dataset_data(node, None, has_facets)
        return Dataset(name=name, namespace=namespace, facets=facets)

    def node_to_output_dataset(
        self,
        node: ModelNode,
        has_facets: bool = False,
        adapter_response: Optional[Dict] = None,
    ) -> OutputDataset:
        namespace, name, facets, _ = self.extract_dataset_data(node, None, has_facets)
        if not has_facets:
            return OutputDataset(name=name, namespace=namespace, facets=facets)

        row_count = 0
        byte_count = 0
        if node.catalog_node:
            row_count = (
                get_from_multiple_chains(
                    node.catalog_node,
                    [
                        ["stats", "num_rows", "value"],  # bigquery
                        ["stats", "row_count", "value"],  # snowflake
                        ["stats", "rows", "value"],  # redshift
                    ],
                )
                or 0
            )

            byte_count = (
                get_from_multiple_chains(
                    node.catalog_node,
                    [
                        ["stats", "num_bytes", "value"],  # bigquery
                        ["stats", "bytes", "value"],  # snowflake
                    ],
                )
                or 0
            )

            if self.adapter_type == Adapter.REDSHIFT:
                # size is the number of 1MB blocks
                blocks_count = (
                    get_from_multiple_chains(
                        node.catalog_node,
                        [
                            ["stats", "size", "value"],
                        ],
                    )
                    or 0
                )
                byte_count = byte_count or blocks_count * (2**20)

        if adapter_response:
            row_count = row_count or adapter_response.get("rows_affected", 0)
            byte_count = byte_count or adapter_response.get("bytes_processed", 0)

        row_count = int(row_count)
        byte_count = int(byte_count)
        if not row_count and not byte_count:
            return OutputDataset(name=name, namespace=namespace, facets=facets)

        output_facets: Dict[str, OutputDatasetFacet] = {}
        output_facets[
            "outputStatistics"
        ] = output_statistics_output_dataset.OutputStatisticsOutputDatasetFacet(
            rowCount=row_count if row_count > 0 else None,
            size=byte_count if byte_count > 0 else None,
        )
        return OutputDataset(name=name, namespace=namespace, facets=facets, outputFacets=output_facets)

    def _format_dataset_name(self, database: Optional[str], schema: Optional[str], table: str) -> str:
        return ".".join(list(filter(None, [database, schema, table])))

    def extract_dataset_data(
        self,
        node: ModelNode,
        assertions: Optional[data_quality_assertions_dataset.DataQualityAssertionsDatasetFacet],
        has_facets: bool = False,
    ) -> Tuple[str, str, Dict, Dict]:
        facets: Dict[str, DatasetFacet]
        input_facets: Dict[str, InputDatasetFacet] = {}
        if has_facets:
            facets = {
                "dataSource": datasource_dataset.DatasourceDatasetFacet(
                    name=self.dataset_namespace, uri=self.dataset_namespace
                ),
            }

            documentation = node.metadata_node["description"]
            if documentation:
                facets["documentation"] = documentation_dataset.DocumentationDatasetFacet(
                    description=documentation
                )

            if assertions:
                input_facets["dataQualityAssertions"] = assertions

            if node.catalog_node:
                fields = self.extract_catalog_fields(
                    node.catalog_node["columns"].values(),
                    node.metadata_node["columns"],
                )
            else:
                fields = self.extract_metadata_fields(node.metadata_node["columns"].values())
            if fields:
                facets["schema"] = schema_dataset.SchemaDatasetFacet(fields=fields)
        else:
            facets = {}
        if node.type == "source":
            table = node.metadata_node["name"]
        else:
            table = node.metadata_node["alias"]
        return (
            self.dataset_namespace,
            self._format_dataset_name(
                node.metadata_node["database"],
                node.metadata_node["schema"],
                table,
            ),
            facets,
            input_facets,
        )

    @staticmethod
    def extract_metadata_fields(columns: List[Dict]) -> List[schema_dataset.SchemaDatasetFacetFields]:
        """
        Extract table field info from metadata's node column info
        Should be used only in the lack of catalog's presence, as there's less
        information in metadata file than in catalog.
        """
        fields = []
        for field in columns:
            fields.append(
                schema_dataset.SchemaDatasetFacetFields(
                    name=field["name"],
                    type=field.get("data_type", None),
                    description=field.get("description", None),
                )
            )
        return fields

    @staticmethod
    def extract_catalog_fields(
        columns: List[Dict], metadata_columns: Dict
    ) -> List[schema_dataset.SchemaDatasetFacetFields]:
        """Extract table field info from catalog's node column info"""
        fields = []
        for field in columns:
            name = field["name"]
            type_ = field.get("type", None)
            assert isinstance(type_, str), f"Catalog field {field} type is null"
            description = get_from_nullable_chain(metadata_columns, [name, "description"])
            fields.append(
                schema_dataset.SchemaDatasetFacetFields(name=name, type=type_, description=description)
            )
        return fields

    def extract_adapter_type(self, profile: Dict):
        try:
            self.adapter_type = Adapter[profile["type"].upper()]
        except KeyError:
            raise NotImplementedError(
                f"Only {Adapter.adapters()} adapters are supported right now. Passed {profile['type']}"
            )

    def extract_dataset_namespace(self, profile: Dict):
        self.dataset_namespace = self.extract_namespace(profile)

    def extract_namespace(self, profile: Dict) -> str:
        """Extract namespace from profile's type"""
        if self.adapter_type == Adapter.SNOWFLAKE:
            return f"snowflake://{fix_account_name(profile['account'])}"
        elif self.adapter_type == Adapter.BIGQUERY:
            return "bigquery"
        elif self.adapter_type == Adapter.REDSHIFT:
            return f"redshift://{profile['host']}:{profile['port']}"
        elif self.adapter_type == Adapter.POSTGRES:
            return f"postgres://{profile['host']}:{profile['port']}"
        elif self.adapter_type == Adapter.CLICKHOUSE:
            return f"clickhouse://{profile['host']}:{profile['port']}"
        elif self.adapter_type == Adapter.TRINO:
            return f"trino://{profile['host']}:{profile['port']}"
        elif self.adapter_type == Adapter.DATABRICKS:
            return f"databricks://{profile['host']}"
        elif self.adapter_type == Adapter.SQLSERVER:
            return f"mssql://{profile['server']}:{profile['port']}"
        elif self.adapter_type == Adapter.DREMIO:
            return f"dremio://{profile['software_host']}:{profile['port']}"
        elif self.adapter_type == Adapter.ATHENA:
            return f"awsathena://athena.{profile['region_name']}.amazonaws.com"
        elif self.adapter_type == Adapter.GLUE:
            if "account_id" in profile:
                return f"arn:aws:glue:{profile['region']}:{profile['account_id']}"
            else:
                # derive account_id from role_arn (arn:aws:iam::account_id:role/role_name)
                return f"arn:aws:glue:{profile['region']}:{profile['role_arn'].split(':')[4]}"
        elif self.adapter_type == Adapter.DUCKDB:
            return f"duckdb://{profile['path']}"
        elif self.adapter_type == Adapter.SPARK:
            port = ""

            if "port" in profile:
                port = f":{profile['port']}"
            elif profile["method"] in [
                SparkConnectionMethod.HTTP.value,
                SparkConnectionMethod.ODBC.value,
            ]:
                port = "443"
            elif profile["method"] == SparkConnectionMethod.THRIFT.value:
                port = "10001"

            if profile["method"] in SparkConnectionMethod.methods():
                return f"spark://{profile['host']}{port}"
            else:
                raise NotImplementedError(
                    f"Connection method `{profile['method']}` is not supported for spark adapter."
                )
        else:
            raise NotImplementedError(
                f"Only {Adapter.adapters()} adapters are supported right now. Passed {profile['type']}"
            )

    def extract_dialect(self) -> str | None:
        if not self.adapter_type:
            return None
        return self.adapter_type.value.lower()

    def get_run(
        self, run_id: str, query_id: Optional[str] = None, run_facets: Optional[dict[str, Any]] = None
    ) -> Run:
        if run_facets is None:
            run_facets = {}

        run_facets.update(
            {
                **self.dbt_version_facet(),
                **self.dbt_run_run_facet(),
                **self.processing_engine_facet(),
            }
        )
        if self._dbt_run_metadata:
            run_facets["parent"] = self._dbt_run_metadata.to_openlineage()

        if query_id:
            run_facets["externalQuery"] = external_query_run.ExternalQueryRunFacet(
                externalQueryId=query_id, source=self.dataset_namespace
            )

        return Run(
            runId=run_id,
            facets=run_facets,
        )

    # TODO: remove after deprecation period
    def dbt_version_facet(self) -> dict[str, DbtVersionRunFacet]:
        dbt_version = self.run_metadata.get("dbt_version")
        if not dbt_version:
            return {}

        self.logger.debug(
            "dbt_version facet is deprecated, and will be removed in future versions. "
            "Use processing_engine facet instead."
        )
        return {"dbt_version": DbtVersionRunFacet(version=dbt_version)}

    @abstractmethod
    def dbt_run_run_facet(self) -> dict[str, DbtRunRunFacet]:
        ...

    def processing_engine_facet(self) -> dict[str, processing_engine_run.ProcessingEngineRunFacet]:
        dbt_version = self.run_metadata.get("dbt_version")
        if not dbt_version:
            return {}
        return {
            "processing_engine": processing_engine_run.ProcessingEngineRunFacet(
                name="dbt",
                version=dbt_version,
                openlineageAdapterVersion=openlineage_version,
            )
        }

    @staticmethod
    def get_timings(timings: List[Dict]) -> Tuple[str, str]:
        """Extract timing info from run_result's timing dict"""
        try:
            timing = list(filter(lambda x: x["name"] == "execute", timings))[0]
            return timing["started_at"], timing["completed_at"]
        except IndexError:
            # Run failed: there is no timing data
            timing_str = datetime.datetime.now(datetime.timezone.utc).isoformat()
            return timing_str, timing_str

    @staticmethod
    def removeprefix(string: str, prefix: str) -> str:
        if string.startswith(prefix):
            return string[len(prefix) :]
        else:
            return string[:]

    def get_column_lineage(
        self, namespace: str, compiled_sql: str
    ) -> Optional[column_lineage_dataset.ColumnLineageDatasetFacet]:
        """Parse SQL and extract column-level lineage information

        Args:
            namespace: The namespace for the column lineage
            compiled_sql: The compiled SQL to parse

        Returns:
            ColumnLineageDatasetFacet if lineage can be parsed, None otherwise
        """
        try:
            parsed = parse_sql([compiled_sql])
            fields = {}
            if parsed:
                for cll_item in parsed.column_lineage:
                    fields[cll_item.descendant.name] = column_lineage_dataset.Fields(
                        inputFields=[
                            column_lineage_dataset.InputField(
                                namespace=namespace,
                                name=self._format_dataset_name(
                                    column_meta.origin.database,
                                    column_meta.origin.schema,
                                    column_meta.origin.name,
                                ),
                                field=column_meta.name,
                            )
                            for column_meta in cll_item.lineage
                            if column_meta.origin
                        ],
                    )
            if fields:
                return column_lineage_dataset.ColumnLineageDatasetFacet(fields=fields)
        except Exception as e:
            self.logger.warning(f"Failed to parse column lineage: {e}")
        return None



# Copyright 2018-2025 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0
import pytest
from openlineage.common.provider.dbt.utils import CONSUME_STRUCTURED_LOGS_COMMAND_OPTION
from openlineage.common.utils import (
    IncrementalFileReader,
    add_command_line_arg,
    add_or_replace_command_line_option,
    get_from_nullable_chain,
    parse_multiple_args,
    parse_single_arg,
    remove_command_line_option,
)


def test_nullable_chain_fails():
    x = {"first": {"second": {}}}
    assert get_from_nullable_chain(x, ["first", "second", "third"]) is None


def test_nullable_chain_works():
    x = {"first": {"second": {"third": 42}}}
    assert get_from_nullable_chain(x, ["first", "second", "third"]) == 42

    x = {"first": {"second": {"third": 42, "fourth": {"empty": 56}}}}
    assert get_from_nullable_chain(x, ["first", "second", "third"]) == 42


def test_parse_single_arg_does_not_exist():
    assert parse_single_arg(["dbt", "run"], ["-t", "--target"]) is None
    assert parse_single_arg(["python", "main.py", "--random_arg", "yes"], ["--what"]) is None


def test_parse_single_arg_next():
    assert parse_single_arg(["dbt", "run", "--target", "prod"], ["-t", "--target"]) == "prod"
    assert parse_single_arg(["python", "--random=yes", "--what", "asdf"], ["--what"]) == "asdf"


def test_parse_single_arg_equals():
    assert parse_single_arg(["dbt", "run", "--target=prod"], ["-t", "--target"]) == "prod"
    assert parse_single_arg(["python", "--random=yes", "--what=asdf"], ["--what"]) == "asdf"


def test_parse_single_arg_gets_first_key():
    assert parse_single_arg(["dbt", "run", "--target=prod", "-t=a"], ["-t", "--target"]) == "a"


def test_parse_single_arg_default():
    assert parse_single_arg(["dbt", "run"], ["-t", "--target"]) is None
    assert parse_single_arg(["dbt", "run"], ["-t", "--target"], default="prod") == "prod"


def test_parse_multiple_args():
    assert parse_multiple_args(["dbt", "run", "--foo", "bar"], ["-m", "--model", "--models"]) == []
    assert sorted(
        parse_multiple_args(
            [
                "dbt",
                "run",
                "--foo",
                "bar",
                "--models",
                "model1",
                "model2",
                "-m",
                "model3",
                "--something",
                "else",
                "--model=model4",
            ],
            ["-m", "--model", "--models"],
        )
    ) == ["model1", "model2", "model3", "model4"]


@pytest.mark.parametrize(
    "command_line, arg_name, arg_value, expected_command_line",
    [
        (
            ["dbt", "run", "--select", "orders"],
            "--log-format",
            "json",
            ["dbt", "run", "--select", "orders", "--log-format", "json"],
        ),
        (
            ["dbt", "run", "--select", "orders", "--log-format", "text"],
            "--log-format",
            "json",
            ["dbt", "run", "--select", "orders", "--log-format", "json"],
        ),
    ],
    ids=["add_new_arg", "replace_arg_value"],
)
def test_add_command_line_arg(command_line, arg_name, arg_value, expected_command_line):
    actual_command_line = add_command_line_arg(command_line, arg_name, arg_value)
    assert actual_command_line == expected_command_line


@pytest.mark.parametrize(
    "command_line, option, replace_option, expected_command_line",
    [
        (
            ["dbt", "run", "--select", "orders"],
            "--write-json",
            None,
            ["dbt", "run", "--select", "orders", "--write-json"],
        ),
        (
            ["dbt", "run", "--select", "orders", "--no-write-json"],
            "--write-json",
            "--no-write-json",
            ["dbt", "run", "--select", "orders", "--write-json"],
        ),
        (
            ["dbt", "run", "--select", "orders"],
            "--write-json",
            "--no-write-json",
            ["dbt", "run", "--select", "orders", "--write-json"],
        ),
    ],
    ids=["add_new_option", "replace_option", "replace_non_existing_option"],
)
def test_add_or_replace_command_line_option(command_line, option, replace_option, expected_command_line):
    actual_command_line = add_or_replace_command_line_option(command_line, option, replace_option)
    assert actual_command_line == expected_command_line


@pytest.mark.parametrize(
    "command_line, command_option, expected_command_line",
    [
        (
            ["dbt", "--foo", "run", "--select", "orders"],
            "--foo",
            ["dbt", "run", "--select", "orders"],
        ),
        (
            ["dbt", "run", "--select", "orders"],
            "--bar",
            ["dbt", "run", "--select", "orders"],
        ),
        (
            ["dbt", CONSUME_STRUCTURED_LOGS_COMMAND_OPTION, "run", "--select", "orders"],
            CONSUME_STRUCTURED_LOGS_COMMAND_OPTION,
            ["dbt", "run", "--select", "orders"],
        ),
    ],
    ids=[
        "remove_existing_command_option",
        "remove_absent_command_option",
        "remove_consume_structured_logs_command_option",
    ],
)
def test_remove_command_line_option(command_line, command_option, expected_command_line):
    actual_command_line = remove_command_line_option(command_line, command_option)
    assert actual_command_line == expected_command_line


@pytest.mark.parametrize(
    "incremental_reads, expected_lines",
    [
        (["foo", " bar", " bim\nbuzz", " foo\n"], ["foo bar bim", "buzz foo"]),
        (["foo bar bim\n", "buzz foo\n"], ["foo bar bim", "buzz foo"]),
        (["foo bar bim\n", "buzz foo"], ["foo bar bim"]),
        (["foo", " bar", " bim\nbuzz\nfizz\n", "foo\n"], ["foo bar bim", "buzz", "fizz", "foo"]),
    ],
    ids=[
        "new_line_in_middle",
        "new_line_at_the_end",
        "last_read_has_no_new_line",
        "many_new_lines_in_one_read",
    ],
)
def test_incremental_file_reader(incremental_reads, expected_lines):
    class DummyTextFile:
        """
        This simulates a text file that reads incomplete lines.
        """

        def __init__(self):
            self.incremental_reads = incremental_reads
            self.name = "dummy"
            self.i = 0

        def read(self, *args, **kwargs):
            """
            This simulates incomplete reads.
            """
            next_read = self.incremental_reads[self.i] if self.i < len(incremental_reads) else ""
            self.i += 1
            return next_read

    dummy_text_file = DummyTextFile()
    incremental_reader = IncrementalFileReader(dummy_text_file)
    actual_lines_read = []
    for line in incremental_reader.read_lines(4096):
        actual_lines_read.append(line)

    assert expected_lines == actual_lines_read



# Copyright 2018-2025 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from openlineage.common.provider.dbt.cloud import DbtCloudArtifactProcessor
from openlineage.common.provider.dbt.facets import ParentRunMetadata
from openlineage.common.provider.dbt.local import DbtLocalArtifactProcessor
from openlineage.common.provider.dbt.processor import UnsupportedDbtCommand

__all__ = [
    "DbtLocalArtifactProcessor",
    "DbtCloudArtifactProcessor",
    "ParentRunMetadata",
    "UnsupportedDbtCommand",
]



# Copyright 2018-2025 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from typing import Optional

import attr
from openlineage.client.facet_v2 import (
    BaseFacet,
    parent_run,
)
from openlineage.common.schema import GITHUB_LOCATION


@attr.define
class ParentRunMetadata:
    run_id: str
    job_name: str
    job_namespace: str
    root_parent_job_name: Optional[str] = attr.field(default=None)
    root_parent_job_namespace: Optional[str] = attr.field(default=None)
    root_parent_run_id: Optional[str] = attr.field(default=None)

    def to_openlineage(self) -> parent_run.ParentRunFacet:
        root = None
        if self.root_parent_run_id and self.root_parent_job_namespace and self.root_parent_job_name:
            root = parent_run.Root(
                run=parent_run.RootRun(runId=self.root_parent_run_id),
                job=parent_run.RootJob(
                    namespace=self.root_parent_job_namespace, name=self.root_parent_job_name
                ),
            )

        return parent_run.ParentRunFacet(
            run=parent_run.Run(runId=self.run_id),
            job=parent_run.Job(namespace=self.job_namespace, name=self.job_name),
            root=root,
        )


@attr.define
class DbtVersionRunFacet(BaseFacet):
    version: str

    @staticmethod
    def _get_schema() -> str:
        return GITHUB_LOCATION + "dbt-version-run-facet.json"


@attr.define
class DbtRunRunFacet(BaseFacet):
    invocation_id: str
    project_name: Optional[str] = attr.field(default=None)
    dbt_runtime: Optional[str] = attr.field(default=None)
    project_version: Optional[str] = attr.field(default=None)
    profile_name: Optional[str] = attr.field(default=None)
    account_id: Optional[str] = attr.field(default=None)

    @staticmethod
    def _get_schema() -> str:
        return GITHUB_LOCATION + "dbt-run-run-facet.json"



# Copyright 2018-2025 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0
import logging
import os
from typing import Any, Dict, Generator, List, Optional, TextIO

logger = logging.getLogger(__name__)


def get_from_nullable_chain(source: Any, chain: List[str]) -> Optional[Any]:
    """
    Get object from nested structure of objects, where it's not guaranteed that
    all keys in the nested structure exist.
    Intended to replace chain of `dict.get()` statements.
    Example usage:
    if not job._properties.get('statistics')\
        or not job._properties.get('statistics').get('query')\
        or not job._properties.get('statistics').get('query')\
            .get('referencedTables'):
        return None
    result = job._properties.get('statistics').get('query')\
            .get('referencedTables')
    becomes:
    result = get_from_nullable_chain(properties, ['statistics', 'query', 'queryPlan'])
    if not result:
        return None
    """
    chain.reverse()
    try:
        while chain:
            next_key = chain.pop()
            if isinstance(source, dict):
                source = source.get(next_key)
            else:
                source = getattr(source, next_key)
        return source
    except AttributeError:
        return None


def get_from_multiple_chains(source: Dict[str, Any], chains: List[List[str]]) -> Optional[Any]:
    for chain in chains:
        result = get_from_nullable_chain(source, chain)
        if result:
            return result
    return None


def parse_single_arg(args, keys: List[str], default=None) -> Optional[str]:
    """
    In provided argument list, find first key that has value and return that value.
    Values can be passed either as one argument {key}={value}, or two: {key} {value}
    """
    for key in keys:
        for i, arg in enumerate(args):
            if arg == key and len(args) > i:
                return args[i + 1]
            if arg.startswith(f"{key}="):
                return arg.split("=", 1)[1]
    return default


def parse_multiple_args(args, keys: List[str], default=None) -> List[str]:
    """
    Given a list of arguments that support syntax such as `--key=value1 value2`
    (like `--model`), return a list of values associated with the given keys.
    """
    parsed_values = []

    for key in keys:
        cur_key = None
        cur_value = None

        for arg in args:
            # `--foo bar baz` case
            if arg == key:
                cur_key = arg
            # `--foo=bar` with single value case
            elif arg.startswith(f"{key}="):
                parsed_values.append(arg.split("=", 1)[1])
            elif cur_key:
                # Done the with current key
                if arg.startswith("-") and cur_key:
                    cur_key = None
                    cur_value = None
                # If we reach here, then arg is the next value for the current key
                else:
                    cur_value = arg

            # Add cur_value only if it's set, if cur_key is set and the argument is
            if cur_key and cur_value:
                parsed_values.append(cur_value)

    return parsed_values or default or []


def add_command_line_arg(command_line: List[str], arg_name: str, arg_value: str):
    command_line = list(command_line)
    arg_name_index = None
    try:
        arg_name_index = command_line.index(arg_name)
    except ValueError:
        # arg_name is not in list
        pass

    if arg_name_index is not None:
        if arg_name_index + 1 >= len(command_line):
            command_line.append(arg_value)
        else:
            command_line[arg_name_index + 1] = arg_value
    else:
        command_line.append(arg_name)
        if arg_value:
            command_line.append(arg_value)

    return command_line


def add_command_line_args(command_line: List[str], arg_names: List[str], arg_values: List[str]):
    for i in range(len(arg_names)):
        arg_name = arg_names[i]
        arg_value = arg_values[i]
        command_line = add_command_line_arg(command_line, arg_name, arg_value)
    return command_line


def add_or_replace_command_line_option(
    command_line: List[str], option: str, replace_option: Optional[str] = None
) -> List[str]:
    """
    If replace_option is ignored then the option is simply added
    """
    command_line = list(command_line)
    if replace_option:
        try:
            replace_option_index = command_line.index(replace_option)
            command_line[replace_option_index] = option
        except ValueError:
            command_line.append(option)
    else:
        command_line.append(option)

    return command_line


def has_command_line_option(command_line: List[str], command_option: str) -> bool:
    return command_option in command_line


def remove_command_line_option(
    command_line: List[str], command_option: str, remove_value: bool = False
) -> List[str]:
    if not has_command_line_option(command_line, command_option):
        return command_line

    command_line = list(command_line)
    command_option_index = command_line.index(command_option)

    # Remove the option itself
    command_line.pop(command_option_index)

    # Optionally remove the value if requested and it exists
    if (
        remove_value
        and command_option_index < len(command_line)
        and not command_line[command_option_index].startswith("-")
    ):
        # Remove the value as well
        command_line.pop(command_option_index)

    return command_line


def has_lines(text_file: TextIO):
    current_cursor = text_file.tell()
    lines = text_file.readlines()
    text_file.seek(current_cursor)
    return len(lines) > 0


class IncrementalFileReader:
    def __init__(self, text_file: TextIO):
        """
        dbt process writes to the log file incrementally. Sometimes lines are written incomplete.
        This class is responsible of reading complete lines only and returns them via its methods.
        """
        self.text_file = text_file
        self.incomplete_chunks: List[str] = []
        self.chunk_size = 65536

    def read_lines(self, max_read_size: int) -> Generator[str, Any, None]:
        logger.debug("Reading %d bytes from %s", max_read_size, self.text_file.name)
        read_since_start = 0
        current_read = 0

        while max_read_size > 0:
            to_read = min(max_read_size, self.chunk_size)
            read_data = self.text_file.read(to_read)

            if not read_data:
                break

            max_read_size -= len(read_data)
            read_since_start += len(read_data)
            current_read += len(read_data)

            if current_read >= 100000:
                logger.debug("Read %d bytes. Left to read %d bytes", read_since_start, max_read_size)
                current_read = 0

            separators = []
            pos = 0
            while True:
                sep_pos = read_data.find(os.linesep, pos)
                if sep_pos == -1:
                    break
                separators.append(sep_pos)
                pos = sep_pos + 1

            if separators:
                incomplete_line = "".join(self.incomplete_chunks)
                full_data = incomplete_line + read_data
                lines = []
                start = 0

                for sep_pos in separators:
                    actual_pos = len(incomplete_line) + sep_pos
                    lines.append(full_data[start:actual_pos])
                    start = actual_pos + 1

                yield from lines

                # no sep at the end
                self.incomplete_chunks = [full_data[start:]]
            else:
                # no separators at all - accumulate chunks instead of concatenating strings
                # it's more efficient to do only one large concat at the end when we
                # find out where the separator is
                self.incomplete_chunks.append(read_data)

        logger.debug("Finished reading, read %d bytes", read_since_start)



# Copyright 2018-2025 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0
import datetime
import json
import os
import subprocess
import sys
import time
from collections import defaultdict
from functools import cached_property, wraps
from typing import Dict, Generator, List, NamedTuple, Optional, TextIO

from openlineage.client.event_v2 import Dataset, RunEvent, RunState
from openlineage.client.facet_v2 import (
    data_quality_assertions_dataset as dq,
)
from openlineage.client.facet_v2 import (
    error_message_run,
    external_query_run,
    job_type_job,
    processing_engine_run,
    sql_job,
    tags_run,
)
from openlineage.client.run import InputDataset
from openlineage.client.uuid import generate_new_uuid
from openlineage.common.provider.dbt.facets import DbtRunRunFacet, DbtVersionRunFacet, ParentRunMetadata
from openlineage.common.provider.dbt.local import DbtLocalArtifactProcessor
from openlineage.common.provider.dbt.processor import (
    ModelNode,
    UnsupportedDbtCommand,
)
from openlineage.common.provider.dbt.utils import (
    DBT_LOG_FILE_MAX_BYTES,
    HANDLED_COMMANDS,
    generate_run_event,
    get_dbt_command,
    get_dbt_log_path,
    get_dbt_profiles_dir,
    get_event_timestamp,
    get_job_type,
    get_node_unique_id,
    get_parent_run_metadata,
    is_random_logfile,
)
from openlineage.common.provider.dbt.utils import (
    __version__ as openlineage_version,
)
from openlineage.common.utils import (
    IncrementalFileReader,
    add_command_line_args,
    add_or_replace_command_line_option,
    get_from_nullable_chain,
    has_lines,
)


class ManifestIntegrityResult(NamedTuple):
    """Result of manifest integrity validation.

    Attributes:
        is_valid: bool indicating if manifest is valid
        missing_nodes: list of node IDs that are missing
        missing_parents: list of parent node IDs that are missing
        missing_children: list of child node IDs that are missing
        total_missing: total count of missing nodes
        parent_map_size: total number of entries in parent_map
        available_nodes: total number of available nodes (nodes + sources)
    """

    is_valid: bool
    missing_nodes: List[str]
    missing_parents: List[str]
    missing_children: List[str]
    total_missing: int
    parent_map_size: int
    available_nodes: int


def handle_keyerror(func):
    """
    Handle KeyError exceptions when node_id was not found in the dbt manifest.
    """

    @wraps(func)
    def wrapper(self, node_id: str, *args, **kwargs):
        try:
            return func(self, node_id, *args, **kwargs)
        except KeyError:
            self.logger.warning(
                "Did not find node_id %s in the dbt manifest - did "
                "the manifest change while the job was executing?",
                node_id,
            )
            return None

    return wrapper


class DbtStructuredLogsProcessor(DbtLocalArtifactProcessor):
    should_raise_on_unsupported_command = True

    def __init__(
        self,
        dbt_command_line: List[str],
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.dbt_command_line: List[str] = dbt_command_line
        self.profiles_dir: str = get_dbt_profiles_dir(command=self.dbt_command_line)
        self.dbt_log_file_path: str = get_dbt_log_path(command=self.dbt_command_line)
        self.is_random_logfile: bool = is_random_logfile(command=self.dbt_command_line)
        self.dbt_log_dirname: str = os.path.dirname(self.dbt_log_file_path)
        self.parent_run_metadata: ParentRunMetadata = get_parent_run_metadata()

        self.node_id_to_ol_run_id: Dict[str, str] = defaultdict(lambda: str(generate_new_uuid()))

        # sql query ids are incremented sequentially per node_id
        self.node_id_to_sql_query_id: Dict[str, Dict[str, int]] = defaultdict(lambda: {"next_id": 1})
        self.node_id_to_sql_start_event: Dict[str, RunEvent] = {}

        self.node_id_to_inputs: Dict[str, List[ModelNode]] = {}
        self.node_id_to_output: Dict[str, ModelNode] = {}

        # will be populated when some dbt events are collected
        self._compiled_manifest: Dict = {}
        self._dbt_version: Optional[str] = None
        self._dbt_invocation_id: Optional[str] = None
        self._dbt_log_file: Optional[TextIO] = None
        self.received_dbt_command_completed = False
        self.processed_bytes = 0

        self.dbt_command_return_code = 0

    @cached_property
    def dbt_command(self) -> str:
        return get_dbt_command(self.dbt_command_line)

    @property
    def dbt_version(self) -> Optional[str]:
        """
        Extracted from the first structured log MainReportVersion
        """
        return self._dbt_version

    @property
    def invocation_id(self) -> Optional[str]:
        return self._dbt_invocation_id

    @cached_property
    def catalog(self) -> Optional[Dict]:
        if os.path.isfile(self.catalog_path):
            return self.load_metadata(self.catalog_path, [1], self.logger)
        return None

    @cached_property
    def profile(self):
        profile_dict = self.load_yaml_with_jinja(
            os.path.expanduser(os.path.join(self.profiles_dir, "profiles.yml"))
        )[self.profile_name]
        if not self.target:
            self.target = profile_dict["target"]

        current_profile = profile_dict["outputs"][self.target]
        return current_profile

    @property
    def compiled_manifest(self) -> Dict:
        """
        Manifest is loaded and cached.
        It's loaded when the dbt structured logs are generated and processed.
        """
        if self._compiled_manifest != {}:
            return self._compiled_manifest
        elif os.path.isfile(self.manifest_path):
            self._compiled_manifest = self.load_metadata(self.manifest_path, list(range(2, 13)), self.logger)
            manifest_data = self._validate_manifest_integrity()
            self.logger.debug(manifest_data)
            return self._compiled_manifest
        else:
            return {}

    def parse(self) -> Generator[RunEvent, None, None]:
        """
        This executes the dbt command and parses the structured log events emitted.
        OL events are generated and yielded when relevant.
        """
        if self.dbt_command not in HANDLED_COMMANDS:
            raise UnsupportedDbtCommand(
                f"dbt integration for structured logs doesn't recognize dbt command "
                f"{self.dbt_command_line} - operation should be one of {HANDLED_COMMANDS}"
            )

        self.extract_adapter_type(self.profile)

        self.extract_dataset_namespace(self.profile)

        for line in self._run_dbt_command():
            ol_event = self._parse_structured_log_event(line)
            if ol_event:
                yield ol_event

        if not self.received_dbt_command_completed:
            # We did not receive the CommandCompleted event, so we emit an abort event
            self.logger.debug("CommandCompleted event was not received. ABORT event will be send.")
            ol_event = self._get_dbt_command_abort_event()
            if ol_event:
                yield ol_event

    def _parse_structured_log_event(self, line: str) -> Optional[RunEvent]:
        """
        dbt generates structured events https://docs.getdbt.com/reference/events-logging
        relevant events are consumed to generate OL events.
        They are usually composed of start/finish events:
        1. NodeStart/NodeFinish for node lifecycle events
        2. MainReportVersion/CommandCompleted for dbt command lifecycle events
        3. SQLQuery/SQLQueryStatus/CatchableExceptionOnRun For SQL query lifecycle events
        """
        self.processed_bytes += len(line)
        self.logger.debug(
            "Received event of size %d. Total processed bytes %d", len(line), self.processed_bytes
        )
        try:
            dbt_event = json.loads(line)
        except ValueError:
            # log that can't be consumed
            self.logger.error("The following log is not valid JSON:\n%s", line)
            return None

        dbt_event_name = dbt_event["info"]["name"]
        self.logger.debug("Parsing dbt event: %s", dbt_event_name)
        if dbt_event_name == "MainReportVersion":
            self._dbt_version = dbt_event["data"]["version"][1:]
            self._dbt_invocation_id = dbt_event["info"].get("invocation_id")
            start_event = self._parse_dbt_start_command_event(dbt_event)
            self._setup_dbt_run_metadata(start_event)
            return start_event

        elif dbt_event_name == "CommandCompleted":
            end_event = self._parse_command_completed_event(dbt_event)
            self.received_dbt_command_completed = True
            return end_event

        if get_node_unique_id(dbt_event) is None:
            return None

        ol_event = None

        if dbt_event_name == "NodeStart":
            ol_event = self._parse_node_start_event(dbt_event)
        elif dbt_event_name == "SQLQuery":
            ol_event = self._parse_sql_query_event(dbt_event)
        elif dbt_event_name in ("SQLQueryStatus", "CatchableExceptionOnRun"):
            ol_event = self._parse_sql_query_status_event(dbt_event)
        elif dbt_event_name == "NodeFinished":
            ol_event = self.parse_node_finished_event(dbt_event)

        return ol_event

    def _parse_dbt_start_command_event(self, event) -> RunEvent:
        event_time = get_event_timestamp(event["info"]["ts"])
        run_facets = {
            **self.dbt_version_facet(),
            **self.processing_engine_facet(),
            **self.dbt_run_run_facet(),
        }

        if self.parent_run_metadata:
            run_facets["parent"] = self.parent_run_metadata.to_openlineage()

        start_event_run_id = str(generate_new_uuid())
        job_facets = {
            "jobType": job_type_job.JobTypeJobFacet(
                jobType=get_job_type(event),
                integration="DBT",
                processingType="BATCH",
                producer=self.producer,
            )
        }

        start_event = generate_run_event(
            event_type=RunState.START,
            event_time=event_time,
            run_id=start_event_run_id,
            job_name=self.job_name,
            job_namespace=self.job_namespace,
            job_facets=job_facets,
            run_facets=run_facets,
        )

        return start_event

    def _get_dbt_command_abort_event(self):
        run_facets = {
            **self.dbt_version_facet(),
            **self.processing_engine_facet(),
            **self.dbt_run_run_facet(),
        }

        parent_run_metadata = get_parent_run_metadata()
        if parent_run_metadata:
            run_facets["parent"] = parent_run_metadata.to_openlineage()

        job_facets = {
            "jobType": job_type_job.JobTypeJobFacet(
                jobType="JOB",
                integration="DBT",
                processingType="BATCH",
                producer=self.producer,
            )
        }
        if not self.dbt_run_metadata:
            return None
        return generate_run_event(
            event_type=RunState.ABORT,
            event_time=datetime.datetime.now().isoformat(),  # Current time - no other data source
            run_id=self.dbt_run_metadata.run_id,
            job_name=self.dbt_run_metadata.job_name,
            job_namespace=self.dbt_run_metadata.job_namespace,
            job_facets=job_facets,
            run_facets=run_facets,
        )

    def _parse_node_start_event(self, event: Dict) -> RunEvent:
        node_unique_id = get_node_unique_id(event)
        run_id = self.node_id_to_ol_run_id[node_unique_id]
        node_start_time = get_event_timestamp(event["data"]["node_info"]["node_started_at"])

        run_facets = {
            **self.dbt_version_facet(),
            **self.processing_engine_facet(),
            **self.dbt_run_run_facet(),
            "parent": self.dbt_run_metadata.to_openlineage(),
        }

        # Add tags if they exist for this node
        if tags := self._get_node_tags(node_unique_id):
            run_facets["tags"] = tags_run.TagsRunFacet(
                tags=[tags_run.TagsRunFacetFields(key=tag, value="true", source="DBT") for tag in tags]
            )

        job_name = self._get_job_name(event)
        job_facets = {
            "jobType": job_type_job.JobTypeJobFacet(
                jobType=get_job_type(event),
                integration="DBT",
                processingType="BATCH",
                producer=self.producer,
            )
        }

        inputs = [
            self.node_to_dataset(node=model_input, has_facets=True)
            for model_input in self._get_model_inputs(node_unique_id)
        ]
        outputs = []
        if node := self._get_model_node(node_unique_id):
            outputs = [self.node_to_output_dataset(node=node, has_facets=True)]

        return generate_run_event(
            event_type=RunState.START,
            event_time=node_start_time,
            run_id=run_id,
            run_facets=run_facets,
            job_name=job_name,
            job_namespace=self.job_namespace,
            job_facets=job_facets,
            inputs=inputs,
            outputs=outputs,
        )

    def parse_node_finished_event(self, event) -> RunEvent:
        resource_type = event["data"]["node_info"]["resource_type"]
        node_unique_id = get_node_unique_id(event)
        node_finished_at = get_event_timestamp(event["data"]["node_info"]["node_finished_at"])
        if node_finished_at == "":
            node_finished_at = get_event_timestamp(event["data"]["node_info"]["node_started_at"])
        node_status = event["data"]["node_info"]["node_status"]
        run_id = self.node_id_to_ol_run_id[node_unique_id]

        run_facets = {
            **self.dbt_version_facet(),
            **self.processing_engine_facet(),
            **self.dbt_run_run_facet(),
            "parent": self.dbt_run_metadata.to_openlineage(),
        }

        # Add tags if they exist for this node
        if tags := self._get_node_tags(node_unique_id):
            run_facets["tags"] = tags_run.TagsRunFacet(
                tags=[tags_run.TagsRunFacetFields(key=tag, value="true", source="DBT") for tag in tags]
            )

        job_name = self._get_job_name(event)
        job_facets = {
            "jobType": job_type_job.JobTypeJobFacet(
                jobType=get_job_type(event),
                integration="DBT",
                processingType="BATCH",
                producer=self.producer,
            )
        }

        event_type = RunState.OTHER

        if node_status == "skipped":
            event_type = RunState.ABORT
        elif node_status in ("fail", "error", "runtime error"):
            event_type = RunState.FAIL
            error_message = event["data"]["run_result"]["message"]
            run_facets["errorMessage"] = error_message_run.ErrorMessageRunFacet(
                message=error_message, programmingLanguage="sql"
            )
        elif node_status in ("success", "pass"):
            event_type = RunState.COMPLETE
        else:
            self.logger.info("Node %s has an unknown node status %s", node_unique_id, node_status)

        inputs = [
            self.node_to_dataset(node=model_input, has_facets=True)
            for model_input in self._get_model_inputs(node_unique_id)
        ]
        outputs = []
        if node := self._get_model_node(node_unique_id):
            outputs = [self.node_to_output_dataset(node=node, has_facets=True)]

        if resource_type == "test":
            success = node_status == "pass"
            if assertion := self._get_assertion(node_unique_id, success):
                assertion_facet = dq.DataQualityAssertionsDatasetFacet(assertions=[assertion])
                inputs = []
                if attached_dataset := self._get_attached_dataset(node_unique_id):
                    dataset_facets = attached_dataset.facets
                    dataset_facets["dataQualityAssertions"] = assertion_facet
                    inputs = [
                        Dataset(
                            name=attached_dataset.name,
                            namespace=attached_dataset.namespace,
                            facets=dataset_facets,
                        )
                    ]

        return generate_run_event(
            event_type=event_type,
            event_time=node_finished_at,
            run_id=run_id,
            run_facets=run_facets,
            job_name=job_name,
            job_namespace=self.job_namespace,
            job_facets=job_facets,
            inputs=inputs,
            outputs=outputs,
        )

    @handle_keyerror
    def _get_attached_dataset(self, test_node_id: str) -> Optional[InputDataset]:
        """
        gets the input the data test is attached to.
        Some nodes like tests related to sources have an attached_node to None.
        """
        all_nodes = {**self.compiled_manifest["nodes"], **self.compiled_manifest["sources"]}
        test_node = all_nodes[test_node_id]
        attached_node_id = test_node["attached_node"]
        input_dataset = None
        if attached_node_id:
            attached_model_node = self._get_model_node(attached_node_id)
            input_dataset = self.node_to_dataset(node=attached_model_node, has_facets=True)
        return input_dataset

    @handle_keyerror
    def _get_assertion(self, node_id: str, success: bool) -> Optional[dq.Assertion]:
        manifest_test_node = self.compiled_manifest["nodes"][node_id]
        name = manifest_test_node["test_metadata"]["name"]
        return dq.Assertion(
            assertion=name,
            success=success,
            column=get_from_nullable_chain(manifest_test_node["test_metadata"], ["kwargs", "column_name"]),
        )

    def _parse_sql_query_event(self, event) -> RunEvent:
        node_unique_id = get_node_unique_id(event)
        node_start_run_id = self.node_id_to_ol_run_id[node_unique_id]
        sql_ol_run_id = str(generate_new_uuid())
        sql_start_at = event["info"]["ts"]

        parent_run = ParentRunMetadata(
            run_id=node_start_run_id,
            job_name=self._get_job_name(event),
            job_namespace=self.job_namespace,
            root_parent_run_id=self.get_root_parent_run_id(),
            root_parent_job_namespace=self.get_root_parent_job_namespace(),
            root_parent_job_name=self.get_root_parent_job_name(),
        )

        run_facets = {
            **self.dbt_version_facet(),
            **self.processing_engine_facet(),
            **self.dbt_run_run_facet(),
            "parent": parent_run.to_openlineage(),
        }

        job_facets = {
            "jobType": job_type_job.JobTypeJobFacet(
                jobType=get_job_type(event),
                integration="DBT",
                processingType="BATCH",
                producer=self.producer,
            ),
            "sql": sql_job.SQLJobFacet(query=event["data"]["sql"], dialect=self.extract_dialect()),
        }

        sql_event = generate_run_event(
            event_type=RunState.START,
            event_time=sql_start_at,
            run_id=sql_ol_run_id,
            run_facets=run_facets,
            job_name=self._get_sql_job_name(event),
            job_namespace=self.job_namespace,
            job_facets=job_facets,
        )

        self.node_id_to_sql_start_event[node_unique_id] = sql_event

        return sql_event

    def _parse_sql_query_status_event(self, event):
        """
        If the sql query is successful a SQLQueryStatus is generated by dbt.
        In case of failure a CatchableExceptionOnRun is generated instead
        """
        node_unique_id = get_node_unique_id(event)

        # We are heavily relying here on the fact that sql events for sql executions for
        # given node are run sequentially, as there is no information in a log
        # event that would allow us to distinguish different sql events for a single model
        sql_ol_run_event = self.node_id_to_sql_start_event[node_unique_id]

        dbt_node_type = event["info"]["name"]
        run_state = RunState.OTHER
        event_time = get_event_timestamp(event["info"]["ts"])
        run_facets = sql_ol_run_event.run.facets

        if dbt_node_type == "CatchableExceptionOnRun":
            run_state = RunState.FAIL
            error_message = event["data"]["exc"]
            stacktrace = event["data"]["exc_info"]
            run_facets["errorMessage"] = error_message_run.ErrorMessageRunFacet(
                message=error_message, programmingLanguage="sql", stackTrace=stacktrace
            )
        elif dbt_node_type == "SQLQueryStatus":
            run_state = RunState.COMPLETE

            if query_id := get_from_nullable_chain(event, ["data", "query_id"]):
                run_facets["externalQuery"] = external_query_run.ExternalQueryRunFacet(
                    externalQueryId=query_id, source="source"
                )

        return generate_run_event(
            event_type=run_state,
            event_time=event_time,
            run_id=sql_ol_run_event.run.runId,
            run_facets=run_facets,
            job_name=sql_ol_run_event.job.name,
            job_namespace=sql_ol_run_event.job.namespace,
            job_facets=sql_ol_run_event.job.facets,
        )

    def _parse_command_completed_event(self, event) -> RunEvent:
        success = event["data"]["success"]
        event_time = get_event_timestamp(event["data"]["completed_at"])
        run_facets = {
            **self.dbt_version_facet(),
            **self.processing_engine_facet(),
            **self.dbt_run_run_facet(),
        }

        parent_run_metadata = get_parent_run_metadata()
        if parent_run_metadata:
            run_facets["parent"] = parent_run_metadata.to_openlineage()

        if success:
            run_state = RunState.COMPLETE
        else:
            run_state = RunState.FAIL
            error_message = event["info"]["msg"]
            run_facets["errorMessage"] = error_message_run.ErrorMessageRunFacet(
                message=error_message, programmingLanguage="sql"
            )

        job_facets = {
            "jobType": job_type_job.JobTypeJobFacet(
                jobType=get_job_type(event),
                integration="DBT",
                processingType="BATCH",
                producer=self.producer,
            )
        }

        return generate_run_event(
            event_type=run_state,
            event_time=event_time,
            run_id=self.dbt_run_metadata.run_id,
            job_name=self.dbt_run_metadata.job_name,
            job_namespace=self.dbt_run_metadata.job_namespace,
            job_facets=job_facets,
            run_facets=run_facets,
        )

    def _setup_dbt_run_metadata(self, start_event):
        """
        The dbt command defines the parent run metadata of
        all subsequent OL events (NodeStart, NodeFinish ...)
        """
        root_parent_run_id = start_event.run.runId
        root_parent_job_name = start_event.job.name
        root_parent_job_namespace = start_event.job.namespace
        if start_event.run.facets.get("parent"):
            parent_facet = start_event.run.facets.get("parent")
            if hasattr(parent_facet, "root"):
                root_parent_run_id = parent_facet.root.run.runId
                root_parent_job_name = parent_facet.root.job.name
                root_parent_job_namespace = parent_facet.root.job.namespace
            else:
                root_parent_run_id = parent_facet.run.runId
                root_parent_job_name = parent_facet.job.name
                root_parent_job_namespace = parent_facet.job.namespace
        self.dbt_run_metadata = ParentRunMetadata(
            run_id=start_event.run.runId,
            job_name=start_event.job.name,
            job_namespace=start_event.job.namespace,
            root_parent_run_id=root_parent_run_id,
            root_parent_job_name=root_parent_job_name,
            root_parent_job_namespace=root_parent_job_namespace,
        )

    # TODO: remove after deprecation period
    def dbt_version_facet(self) -> Dict[str, DbtVersionRunFacet]:
        if not self.dbt_version:
            return {}
        self.logger.debug(
            "dbt_version facet is deprecated, and will be removed in future versions. "
            "Use processing_engine facet instead."
        )
        return {"dbt_version": DbtVersionRunFacet(version=self.dbt_version)}

    def dbt_run_run_facet(self) -> Dict[str, DbtRunRunFacet]:
        if not self.invocation_id:
            return {}
        return {
            "dbt_run": DbtRunRunFacet(
                invocation_id=self.invocation_id,
                project_name=self.project_name,
                project_version=self.project_version,
                profile_name=self.profile_name,
                dbt_runtime="core",
            )
        }

    def processing_engine_facet(self) -> Dict[str, processing_engine_run.ProcessingEngineRunFacet]:
        if not self.dbt_version:
            return {}
        return {
            "processing_engine": processing_engine_run.ProcessingEngineRunFacet(
                name="dbt",
                version=self.dbt_version,
                openlineageAdapterVersion=openlineage_version,
            )
        }

    def _get_sql_query_id(self, timestamp: str, node_id: str) -> int:
        """
        Not all adapters have the sql id defined in their dbt event.
        A node is executed by a single thread which means that sql queries of a single node are executed
        sequentially and their status is also reported sequentially.
        This function gives us a surrogate query id. It's auto-incremented.
        It gives the same id for the couple node_id, timestamp.
        """
        ids = self.node_id_to_sql_query_id[node_id]
        if timestamp not in ids:
            next_id = ids["next_id"]
            ids[timestamp] = next_id
            ids["next_id"] += 1

        return ids[timestamp]

    def _get_sql_job_name(self, event) -> str:
        """
        The name of the sql job is as follows
        {node_job_name}.sql.{incremental_id}
        """
        timestamp_str = event["info"]["ts"]
        node_unique_id = get_node_unique_id(event)
        query_id = self._get_sql_query_id(timestamp_str, node_unique_id)
        job_name = f"{node_unique_id}.sql.{query_id}"

        return job_name

    def _get_job_name(self, event) -> str:
        """
        The job name of models, snapshots, tests and others.
        """
        return get_node_unique_id(event)

    def _run_dbt_command(self) -> Generator[str, None, None]:
        """
        This executes the dbt command.
        Logs from the dbt log file are consumed in json format.
        """
        dbt_command_line = add_command_line_args(
            self.dbt_command_line,
            arg_names=["--log-format-file", "--log-level-file", "--log-path", "--log-file-max-bytes"],
            arg_values=["json", "debug", self.dbt_log_dirname, DBT_LOG_FILE_MAX_BYTES],
        )
        dbt_command_line = add_or_replace_command_line_option(
            dbt_command_line, option="--write-json", replace_option="--no-write-json"
        )

        self._open_dbt_log_file()
        self.logger.debug("Opened log file %s", self.dbt_log_file_path)

        incremental_reader = IncrementalFileReader(self._dbt_log_file)
        last_size = os.stat(self.dbt_log_file_path).st_size
        self.logger.debug("Running dbt command: %s", " ".join(dbt_command_line))

        process = subprocess.Popen(dbt_command_line, stdout=sys.stdout, stderr=sys.stderr, text=True)
        parse_manifest = True
        last_log = datetime.datetime.now()

        try:
            while process.poll() is None:
                if (datetime.datetime.now() - last_log) >= datetime.timedelta(seconds=10):
                    self.logger.debug("dbt process is still running: waiting for logs to appear")
                    last_log = datetime.datetime.now()
                if parse_manifest and has_lines(self._dbt_log_file) > 0:
                    # Load the manifest as soon as it exists
                    self.compiled_manifest
                    parse_manifest = False
                    self.logger.debug("Parsed manifest file")

                current_size = os.stat(self.dbt_log_file_path).st_size
                if current_size > last_size:
                    self.logger.debug(
                        "Current size: %d, last size: %d, to read: %d",
                        current_size,
                        last_size,
                        current_size - last_size,
                    )
                    yield from incremental_reader.read_lines(current_size - last_size)
                    last_size = current_size
                time.sleep(0.01)

            self.logger.debug("dbt process has exited. Waiting for logs to be flushed")
            max_loops = 10
            i = 0
            while (
                self._dbt_log_file is not None and not self.received_dbt_command_completed and i < max_loops
            ):
                i += 1
                current_size = os.stat(self.dbt_log_file_path).st_size
                if current_size > last_size:
                    self.logger.debug(
                        "Current size: %d, last size: %d, to read: %d",
                        current_size,
                        last_size,
                        current_size - last_size,
                    )
                    yield from incremental_reader.read_lines(current_size - last_size)
                    last_size = current_size
            self.logger.debug("Finished waiting for logs to be flushed")

        except Exception:
            self.logger.exception(
                "An exception occurred in OpenLineage dbt wrapper. "
                "It should not affect underlying dbt execution that is still running."
            )
            last_log = datetime.datetime.now()

            while process.poll() is None:
                if (datetime.datetime.now() - last_log) >= datetime.timedelta(seconds=10):
                    self.logger.debug("dbt process is still running. Further logs won't be processed")
                    last_log = datetime.datetime.now()
                time.sleep(0.1)
                pass
            self.logger.debug("dbt process has finished")
        finally:
            if self._dbt_log_file is not None:
                self._dbt_log_file.close()
                if self.is_random_logfile:
                    try:
                        self.logger.debug("Removing log file: %s", self.dbt_log_file_path)
                        os.remove(self.dbt_log_file_path)
                    except Exception:
                        self.logger.warning(
                            "Failed to remove log file: %s. Leftover logs can consume disk space.",
                            self.dbt_log_file_path,
                        )
            self.dbt_command_return_code = process.returncode

    def _open_dbt_log_file(self):
        """
        If the log file already exists when the dbt command is executed logs are appended.
        This reads all the lines on first (and only) open to get rid of those previous lines.
        If the file doesn't exist it creates an empty one
        """
        if self._dbt_log_file is not None:
            return

        if not os.path.exists(self.dbt_log_file_path):
            logs_directory = os.path.dirname(self.dbt_log_file_path)
            os.makedirs(logs_directory, exist_ok=True)
            with open(self.dbt_log_file_path, "w"):
                pass

        self._dbt_log_file = open(self.dbt_log_file_path)
        while self._dbt_log_file.readlines():
            pass

    @handle_keyerror
    def _get_model_node(self, node_id) -> Optional[ModelNode]:
        """
        Builds a ModelNode of a given node_id
        """
        if node_id in self.compiled_manifest["nodes"]:
            node_type = "model"
            manifest_node = self.compiled_manifest["nodes"][node_id]
        elif node_id in self.compiled_manifest["sources"]:
            node_type = "source"
            manifest_node = self.compiled_manifest["sources"][node_id]
        else:
            raise RuntimeError(f"{node_id} not found in nodes or sources")
        catalog_node = get_from_nullable_chain(self.catalog, ["nodes", node_id])
        return ModelNode(type=node_type, metadata_node=manifest_node, catalog_node=catalog_node)

    @handle_keyerror
    def _get_node_tags(self, node_id: str) -> List[str]:
        """
        Extract tags from a dbt node in the compiled manifest
        """
        all_nodes = {**self.compiled_manifest["nodes"], **self.compiled_manifest["sources"]}
        manifest_node = all_nodes[node_id]
        return manifest_node.get("tags", [])

    def _get_model_inputs(self, node_id) -> List[ModelNode]:
        """
        Builds the model's upstream inputs
        """
        if node_id in self.node_id_to_inputs:
            return self.node_id_to_inputs[node_id]

        try:
            input_node_ids = self.compiled_manifest["parent_map"][node_id]
            inputs = [node for n in input_node_ids if (node := self._get_model_node(n)) is not None]
            self.node_id_to_inputs[node_id] = inputs
            return inputs
        except KeyError:
            self.logger.warning(
                "Did not find node_id %s in the dbt manifest - did "
                "the manifest change while the job was executing?",
                node_id,
            )
        return []

    def _validate_manifest_integrity(self) -> ManifestIntegrityResult:
        """
        Performs metadata integrity checks on the compiled manifest.

        Validates that all nodes referenced in parent_map exist in the manifest's
        node definitions (either in 'nodes' or 'sources' sections).

        Returns:
            ManifestIntegrityResult: Validation results with integrity check information.
        """
        if not self._compiled_manifest:
            return ManifestIntegrityResult(
                is_valid=True,
                missing_nodes=[],
                missing_parents=[],
                missing_children=[],
                total_missing=0,
                parent_map_size=0,
                available_nodes=0,
            )

        parent_map = self._compiled_manifest.get("parent_map", {})
        all_nodes = {**self._compiled_manifest.get("nodes", {}), **self._compiled_manifest.get("sources", {})}

        missing_nodes = []
        missing_parents = []
        missing_children = []

        for parent_node_id, child_node_ids in parent_map.items():
            # Check that the parent node exists
            if parent_node_id not in all_nodes:
                missing_nodes.append(parent_node_id)
                missing_parents.append(parent_node_id)

            # Check that all child nodes exist
            for child_node_id in child_node_ids:
                if child_node_id not in all_nodes:
                    missing_nodes.append(child_node_id)
                    missing_children.append(child_node_id)

        # Remove duplicates while preserving order
        missing_nodes = list(dict.fromkeys(missing_nodes))
        missing_parents = list(dict.fromkeys(missing_parents))
        missing_children = list(dict.fromkeys(missing_children))

        if missing_nodes:
            self.logger.warning(
                "Manifest integrity check failed: Found %d nodes in parent_map that don't exist "
                "in manifest node definitions, printing first %d of them: %s",
                len(missing_nodes),
                min(10, len(missing_nodes)),
                missing_nodes[: min(10, len(missing_nodes))],  # Log first 10 to avoid overwhelming logs
            )

        return ManifestIntegrityResult(
            is_valid=len(missing_nodes) == 0,
            missing_nodes=missing_nodes,
            missing_parents=missing_parents,
            missing_children=missing_children,
            total_missing=len(missing_nodes),
            parent_map_size=len(parent_map),
            available_nodes=len(all_nodes),
        )

    def get_dbt_metadata(
        self,
    ):
        """
        Replaced by properties
        """
        pass

    def get_root_parent_run_id(self):
        if self.parent_run_metadata is not None:
            return self.parent_run_metadata.root_parent_run_id
        if self.dbt_run_metadata is not None:
            return self.dbt_run_metadata.root_parent_run_id
        return None

    def get_root_parent_job_namespace(self):
        if self.parent_run_metadata is not None:
            return self.parent_run_metadata.root_parent_job_namespace
        if self.dbt_run_metadata is not None:
            return self.dbt_run_metadata.root_parent_job_namespace
        return None

    def get_root_parent_job_name(self):
        if self.parent_run_metadata is not None:
            return self.parent_run_metadata.root_parent_job_name
        if self.dbt_run_metadata is not None:
            return self.dbt_run_metadata.root_parent_job_name
        return None
